(*************************************************************)
(* Grammar for Bare language - Definition v3 pages 8,9,70,71 *)
(* modified to have ident in place of con  var and excon     *)
(*************************************************************)

(*
$File: Common/DEC_GRAMMAR.sml $
$Date: 1992/04/07 14:54:08 $
$Revision: 1.12 $
$Locker:  $
*)

(*$DEC_GRAMMAR *)

signature DEC_GRAMMAR =
sig
  type lab        (* labels *)
   and scon       (* special constants *)
   and con        (* constructors *)
   and id         (* identifiers - variables or constructors *)
   and longid     (* long identifiers - variables or constructors *)
   and tyvar      (* type variables *)
   and tycon      (* type constructors *)
   and longtycon  (* long type constructors *)
   and excon      (* exception constructors *)
   and longstrid  (* structure identifiers *)
(* PVADD *)
(**) and strid      (* structure identifier *)
(* need this to do equality on longstrid *)

  type info       (* info about the position in the source text, errors etc *)

  type PostInfo

  val emptyInfo : PostInfo

  datatype 'a op_opt = OP_OPT of 'a * bool
  datatype 'a WithInfo = WITH_INFO of info * 'a

  datatype atexp =
	SCONatexp of info * scon |         
	IDENTatexp of info * longid op_opt |
	RECORDatexp of info * exprow Option |
	LETatexp of info * dec * exp |
	PARatexp of info * exp

  and exprow =
	EXPROW of info * lab * exp * exprow Option

  and exp =
(*EML*) EXISTSexp of info * match |
(*EML*) FORALLexp of info * match |
(*EML*) TERMINATESexp of info * atexp |
	ATEXPexp of info * atexp |
	APPexp of info * exp * atexp |
	TYPEDexp of info * exp * ty |
	HANDLEexp of info * exp * match |
	RAISEexp of info * exp |
	FNexp of info * match |
	UNRES_INFIXexp of info * atexp list
      
  and match =
        MATCH of info * mrule * match Option

  and mrule =
        MRULE of info * pat * exp

  and dec = 
(*EML*) EQTYPEdec of info * typbind |
(*EML*) SPANTYPEdec of info * datbind |
(*EML*) AXIOMdec of info * ax | 
	VALdec of info * valbind |
	UNRES_FUNdec of info * FValBind |
		(* TEMPORARY: removed when resolving infixes after parsing. *)
	TYPEdec of info * typbind |
	DATATYPEdec of info * datbind |
	ABSTYPEdec of info * datbind * dec |
	EXCEPTIONdec of info * exbind |
	LOCALdec of info * dec * dec |
	OPENdec of info * longstrid WithInfo list |
	SEQdec of info * dec * dec |
	INFIXdec of info * int Option * id list |
	INFIXRdec of info * int Option * id list |
	NONFIXdec of info * id list |
	EMPTYdec of info

  and valbind =
(*EML*) UNDEFvalbind of info * pat * (*?*) valbind Option |
	PLAINvalbind of info * pat * exp * valbind Option |
	RECvalbind of info * valbind

  and FValBind = FVALBIND of info * FClause * FValBind Option
  and FClause  = FCLAUSE of info * atpat list * ty Option * exp * FClause Option |
                 UNDEFFClause of info * atpat list * ty Option * ty Option * FClause Option

  and typbind =
(*EML*) UNDEFtypbind of info * tyvar list * tycon * (*?*) typbind Option |
        TYPBIND of info * tyvar list * tycon * ty * typbind Option

  and datbind =
        DATBIND of info * tyvar list * tycon * conbind * datbind Option

  and conbind =
        CONBIND of info * con op_opt * ty Option * conbind Option

  and exbind =
        EXBIND of info * excon op_opt * ty Option * exbind Option |
        EXEQUAL of info * excon op_opt * longid op_opt * exbind Option

  and atpat =
        WILDCARDatpat of info |
	SCONatpat of info * scon |
	LONGIDatpat of info * longid op_opt |
	RECORDatpat of info * patrow Option |
	PARatpat of info * pat

  and patrow =
        DOTDOTDOT of info |
        PATROW of info * lab * pat * patrow Option

  and pat =
        ATPATpat of info * atpat |
        CONSpat of info * longid op_opt * atpat |
        TYPEDpat of info * pat * ty |
        LAYEREDpat of info * id op_opt * ty Option * pat |
	UNRES_INFIXpat of info * atpat list

  and ty =
        TYVARty of info * tyvar |
        RECORDty of info * tyrow Option |
        CONty of info * ty list * longtycon |
        FNty of info * ty * ty |
        PARty of info * ty

  and tyrow =
        TYROW of info * lab * ty * tyrow Option

(*EML*)
  and ax = AXIOMS of info * exp * ax Option

(* PVADD *)
(**)  val eq_option: (('a * 'a) -> bool) -> (('a Option) * ('a Option)) -> bool
(**)  val eq_list: (('a * 'a) -> bool) -> (('a list) * ('a list)) -> bool
(**)  val eq_int: int * int -> bool
(**)  val eq_string: string * string -> bool
(**)  val eq_real: real * real -> bool

(**)  val eq_lab: lab * lab -> bool
(**)  val eq_scon: scon * scon -> bool
(**)  val eq_con: con * con -> bool
(**)  val eq_id: id * id -> bool
(**)  val eq_longid: longid * longid -> bool
(**)  val eq_tyvar: tyvar * tyvar -> bool
(**)  val eq_tycon: tycon * tycon -> bool
(**)  val eq_longtycon: longtycon * longtycon -> bool
(**)  val eq_excon: excon * excon -> bool
(**)  val eq_strid: strid * strid -> bool
(**)  val eq_longstrid: longstrid * longstrid -> bool

(**)  val eq_op_opt: (('a * 'a) -> bool) -> (('a op_opt) * ('a op_opt)) -> bool
(**)  val eq_with_info: (('a * 'a) -> bool) -> (('a WithInfo) * ('a WithInfo)) -> bool
(**)  val eq_atexp: atexp * atexp -> bool
(**)  val eq_exprow: exprow * exprow -> bool
(**)  val eq_exp: exp * exp -> bool
(**)  val eq_match: match * match -> bool
(**)  val eq_mrule: mrule * mrule -> bool
(**)  val eq_dec: dec * dec -> bool
(**)  val eq_valbind: valbind * valbind -> bool
(**)  val eq_fvalbind: FValBind * FValBind -> bool
(**)  val eq_fclause: FClause * FClause -> bool
(**)  val eq_typbind: typbind * typbind -> bool
(**)  val eq_datbind: datbind * datbind -> bool
(**)  val eq_conbind: conbind * conbind -> bool
(**)  val eq_exbind: exbind * exbind -> bool
(**)  val eq_atpat: atpat * atpat -> bool
(**)  val eq_patrow: patrow * patrow -> bool
(**)  val eq_pat: pat * pat -> bool
(**)  val eq_ty: ty * ty -> bool
(**)  val eq_tyrow: tyrow * tyrow -> bool
(**)  val eq_ax: ax * ax -> bool

end;

