(* Error information *)

(*
$File: Common/ERROR_INFO.sml $
$Date: 1992/04/07 14:54:48 $
$Revision: 1.15 $
$Locker:  $
*)

(*$ERROR_INFO*)

signature ERROR_INFO =
  sig
    type Type and TyVar and TyName and TypeFcn
    type id and longid and tycon and longtycon and excon
    type sigid and strid and longstrid and funid

    datatype Tail = TYCON of tycon	(* Tail end for Covers errors. *)
      		  | STRID of strid

    datatype info =
     (* Core errors: *)
	UNIFICATION of Type * Type
      | LOOKUP_LONGID of longid
      | LOOKUP_LONGTYCON of longtycon
      | NOTCONSTYPE of Type
      | QUALIFIED_ID of longid
      | FREE_TYVARS of TyVar list
      | WRONG_ARITY of {expected: int, actual: int}
      | NOTALLOWEDINREC

     (* Extended ML Equality Errors *)
      | ADMITS_EQUALITY
      | DENIES_EQUALITY
      | PREDICATE of Type
      | LOGEXP of Type

     (* General module errors: *)
      | LOOKUP_SIGID of sigid
      | LOOKUP_LONGSTRID of longstrid
      | LOOKUP_FUNID of funid
      | ASSEMBLY_COVER of strid list * Tail
      | NOTTYPEEXPLICIT
      | NOTEQPRINCIPAL of longtycon list

     (* Signature matching errors: *)
      | S_RIGIDSTRCLASH of longstrid Option
      | MISSINGSTR  of longstrid
      | MISSINGTYPE of longtycon
      | S_CONFLICTINGARITY of longtycon * (TyName * TypeFcn)
      | CONFLICTINGEQUALITY of longtycon * (TyName * TypeFcn)
      | MISSINGVAR of strid list * id
      | MISSINGEXC of strid list * excon
      | S_RIGIDTYCLASH of longtycon
      | S_CONFLICTING_DOMCE of longtycon
      | NOTYENRICHMENT of strid list * id
      | EXCNOTEQUAL of strid list * excon * (Type * Type)

     (* Module unification errors: *)
      | U_RIGIDSTRCLASH of longstrid * longstrid
      | CYCLE of longstrid
      | U_RIGIDTYCLASH of longtycon * longtycon
      | TYPESTRILLFORMEDNESS of longtycon * longtycon
      | U_CONFLICTING_DOMCE of longtycon * longtycon
      | U_CONFLICTINGARITY of longtycon * longtycon
      | RIGIDTYFUNEQERROR of longtycon * longtycon

    type Report
    val reportInfo: info -> Report
  end;
