(* Dynamic values and enviroments - Definition v3 page 46 *)

(*
$File: Interpreter/CoreDynObject.sml $
$Date: 1992/04/29 16:38:56 $
$Revision: 1.4 $
$Locker:  $
*)

(*$CoreDynObject:
	LAB CON SPECIAL_VALUE VAR EXCON EXNAME STORE STRID BASIC_VALUE
	FINMAP SORTED_FINMAP PRETTYPRINT CRASH CORE_DYNOBJECT DEC_GRAMMAR
 *)

functor CoreDynObject(structure Lab: LAB
		      structure Con: CON
		      structure SpecialValue: SPECIAL_VALUE
		      structure Var: VAR
                        sharing type Var.id = Con.id 

		      structure Excon: EXCON
		      structure ExName: EXNAME

		      structure StrId: STRID
			sharing type Var.strid = Excon.strid = StrId.strid

		      structure Grammar: DEC_GRAMMAR
                        sharing type Excon.excon = Grammar.excon
                            and type StrId.longstrid = Grammar.longstrid

		      structure BasicValue: BASIC_VALUE

		      structure FinMap: FINMAP
		      structure SortedFinMap: SORTED_FINMAP
		      
		      structure PP: PRETTYPRINT
			sharing type FinMap.StringTree
				     = SortedFinMap.StringTree
				     = PP.StringTree

		      structure Crash: CRASH
		     ): CORE_DYNOBJECT =
  struct

   (* Copying of arguments to result structure *)

    structure Con     = Con
    structure Lab     = Lab
    structure Var     = Var
    structure Excon   = Excon
    structure ExName  = ExName
    structure Grammar = Grammar

    type SVal   = SpecialValue.SVal
    type scon   = SpecialValue.scon
    type BasVal = BasicValue.BasVal
    type strid = StrId.strid
    type longstrid = StrId.longstrid

    type ('a, 'b) map = ('a, 'b) SortedFinMap.map
				(* We export the sorted map type because
				   datatype Val is in our signature. *)

    infix plus
    val op plus = FinMap.plus

    type StringTree = PP.StringTree

    type Address = unit ref

    datatype Val = Sval       of SVal
		 | BASval     of BasVal
		 | CON0val    of Con.con
		 | CON1val    of Con.con * Val
		 | RECORDval  of (Lab.lab, Val) SortedFinMap.map
		 | EXNAME0val of ExName.ExName
		 | EXNAME1val of ExName.ExName * Val
		 | CLOSUREval of Closure
		 | ADDRval    of Address
                 | INCOMPLETEval

	(* These are abstract to the outside world: *)
	 and Closure  = CLOSURE  of Grammar.match * Env * VarEnv
	 and Env      = ENV      of StrEnv * VarEnv * ExConEnv
	 and VarEnv   = VARENV   of (Var.var, Val) FinMap.map
	 and ExConEnv = EXCONENV of (Excon.excon, ExName.ExName) FinMap.map
	 and StrEnv   = STRENV   of (strid, Env) FinMap.map
                      | INCOMPLETEstruct

    val unit = RECORDval SortedFinMap.empty

   (* We can't make Store() an independent functor easily because Address
      and Val are interrelated. *)

    structure Store: STORE =
      struct
	type Address = Address
	type Val = Val

	val store = ref(FinMap.empty: (Address, Val) FinMap.map)

	val unique = ref
	fun add(addr, v) = (store := FinMap.add(addr, v, !store); unit)

	fun retrieve addr =
	  case FinMap.lookup (!store) addr
	    of Some v => v
	     | None => Crash.impossible "CoreDynObject.Store.retrieve"
      end

    fun pair(v1, v2) =
      let
	val m0 = SortedFinMap.empty
	val m1 = SortedFinMap.add (Lab.<) (Lab.mk_IntegerLab 1, v1, m0)
	val m2 = SortedFinMap.add (Lab.<) (Lab.mk_IntegerLab 2, v2, m1)
      in
	RECORDval m2
      end

    fun mustFind (fname, pr) (m, x) =
      case FinMap.lookup m x
	of Some y => y
	 | None => Crash.impossible("CoreDynObject." ^ fname ^ "(" ^ pr x ^ ")")


    fun mkSValSCon sc = SpecialValue.mkSValSCon sc

    val mkClosure = CLOSURE
    fun unClosure(CLOSURE c) = c

   (* for environments *)

   (* Empty enviroments *)

    val emptySE = STRENV FinMap.empty

    val emptyVE = VARENV FinMap.empty

    val emptyEE = EXCONENV FinMap.empty

   (* for variable environments *)

    fun singleVE(v, value) = VARENV(FinMap.singleton(v, value))

    fun singleEE(e, name)  = EXCONENV(FinMap.singleton(e, name))

    val initialVE = singleVE(Var.var_PRIM, BASval BasicValue.prim)

    fun VE_plus_VE(VARENV m1, VARENV m2) = VARENV(m1 plus m2)

    fun domVE(VARENV m) = FinMap.dom m

    fun trimVE(VARENV m, f) = VARENV(FinMap.filter (fn (x, _) => f x) m)

    val mkEnv = ENV

    val emptyE = ENV(emptySE, emptyVE, emptyEE)

    val initialE = ENV(emptySE, initialVE, emptyEE)

    fun VE_in_E VE = ENV(emptySE, VE, emptyEE)

    fun VE_of_E(ENV(_, VE, _)) = VE

    fun EE_in_E EE = ENV(emptySE, emptyVE, EE)

    fun EE_of_E(ENV(_, _, EE)) = EE

    fun EE_plus_EE(EXCONENV m1, EXCONENV m2) = EXCONENV(m1 plus m2)

    fun domEE(EXCONENV m) = FinMap.dom m

    fun trimEE(EXCONENV m, f) = EXCONENV(FinMap.filter (fn (x, _) => f x) m)

    val singleSE = STRENV o FinMap.singleton

    fun SE_plus_SE(STRENV m1, STRENV m2) = STRENV(m1 plus m2)

    fun SE_in_E SE = ENV(SE, emptyVE, emptyEE)

    fun SE_of_E(ENV(SE, _, _)) = SE

    fun E_plus_SE(ENV(SE1, VE, EE), SE2) =
      ENV(SE_plus_SE(SE1, SE2), VE, EE)

    fun E_plus_E(ENV(SE1, VE1, EE1), ENV(SE2, VE2, EE2)) =
      ENV(SE_plus_SE(SE1, SE2), VE_plus_VE(VE1, VE2), EE_plus_EE(EE1, EE2))

    fun domSE(STRENV m) = FinMap.dom m

    fun trimSE(STRENV m, f) = STRENV(FinMap.filter (fn (x, _) => f x) m)

    fun mapSE(STRENV m, f) = STRENV(FinMap.ComposeMap f m)

   (* Unfolding of recursive closures *)

    fun Rec(VE as VARENV m) =
      let
	fun unfold(CLOSUREval(CLOSURE(match, E', VE''))) =
	      CLOSUREval(CLOSURE(match, E', VE))

	  | unfold v = v
      in
	VARENV(FinMap.composemap unfold m)
      end

    (******** Lookup functions ********)

    fun lookup_StrId(ENV(STRENV m, _, _), strid) =
      mustFind ("lookup_StrId", StrId.pr_StrId) (m, strid)

    fun lookup_Var(ENV(_, VARENV m, _), var) =
      mustFind ("lookup_Var", Var.pr_var) (m, var)

    fun lookup_Excon(ENV(_, _, EXCONENV m), excon) =
      mustFind ("lookup_Excon", Excon.pr_excon) (m, excon)


   (* Generic function to traverse a path of strid's in a qualified ID. *)

    local
      fun pathLookup(E, (strids, id: 'a), f: Env * 'a -> 'b): 'b =
	case strids
	  of nil => f(E, id)

	   | (this :: rest) =>
	       pathLookup(lookup_StrId(E, this), (rest, id), f)
    in
      fun lookup_LongStrId(E, longstrid) =
	pathLookup(E, StrId.explode_longstrid longstrid, lookup_StrId)

      fun lookup_LongVar(E, longvar) =
	pathLookup(E, Var.decompose longvar, lookup_Var)

      fun lookup_LongExcon(E, longexcon) =
	pathLookup(E, Excon.decompose longexcon, lookup_Excon)
    end

   (* exception packets. *)

    datatype Pack = EXNAME0pack of ExName.ExName
		  | EXNAME1pack of ExName.ExName * Val

    exception EXCEPTION of Pack

   (* printing. *)

    val layoutLab = PP.layoutAtom Lab.pr_Lab
    val layoutVar = PP.layoutAtom Var.pr_var
    val layoutStrId = PP.layoutAtom StrId.pr_StrId
    val layoutExcon = PP.layoutAtom Excon.pr_excon
    val layoutExName = PP.layoutAtom ExName.pr_ExName

    fun layoutVal(Sval sval) = PP.layoutAtom SpecialValue.pr_SVal sval
      | layoutVal(BASval bv) = PP.LEAF "fn"
      | layoutVal(CON0val c) = PP.layoutAtom Con.pr_con c

      | layoutVal(CON1val(c, v)) =
	  PP.NODE{start="", finish=")", indent=1,
		  children=[PP.layoutAtom Con.pr_con c, layoutVal v],
		  childsep=PP.LEFT "("
		 }

      | layoutVal(CLOSUREval _) = PP.LEAF "fn"

      | layoutVal(EXNAME0val exname) =
	  layoutExName exname

      | layoutVal(EXNAME1val(exname, v)) =
	  PP.NODE{start=ExName.pr_ExName exname ^ "(", finish=")", indent=3,
		  children=[layoutVal v], childsep=PP.NONE
		 }

     (* For records, we'll see if they look like tuples. Be careful of
        {1=x} records which are not treated this way. *)

      | layoutVal(RECORDval m) =
	  if SortedFinMap.matches (fn (i, lab) => Lab.is_LabN(lab, i+1)) m
	  then
	    case SortedFinMap.rangeSORTED m
	      of [x] =>
		   PP.NODE{start="{1=", finish="}",
			   indent=1, childsep=PP.NONE,
			   children=[layoutVal x]
			  }

	       | xs =>
		   PP.NODE{start="(", finish=")", indent=1,
			   childsep=PP.RIGHT ", ", children=map layoutVal xs
			  }
	  else
	    SortedFinMap.layoutMap {start="{", eq="=", sep=", ", finish="}"}
				   layoutLab layoutVal m

      | layoutVal(ADDRval addr) =
	  PP.NODE{start="(ref ", finish=")", indent=3, childsep=PP.NONE,
		  children=[layoutVal(Store.retrieve addr)]
		 }

      | layoutVal INCOMPLETEval = PP.LEAF "Incomplete Value"

    local
      fun f (layoutL, layoutR) map =
	FinMap.layoutMap {start="{", eq=" -> ", sep=", ", finish="}"}
			 layoutL layoutR map
    in
      fun layoutStrEnv(STRENV strMap) = f (layoutStrId, layoutEnv) strMap
        | layoutStrEnv INCOMPLETEstruct = PP.LEAF "Incomplete structure"

      and layoutEnv(ENV(SE, VARENV varMap, EXCONENV exMap)) =
	PP.NODE{start="(ENV: ", finish=")", indent=3, childsep=PP.RIGHT "; ",
		children=[layoutStrEnv SE,
			  f (layoutVar, layoutVal) varMap,
			  f (layoutExcon, layoutExName) exMap
			 ]
	       }
    end
  end;
