signature PVENV =
  sig
    structure OK:sig
      val parse: unit -> unit
      val elab: unit -> unit
      val eval: unit -> unit

      val parseFile: string -> unit
      val elabFile: string -> unit
      val evalFile: string -> unit

(* PVADD *)
(**)  structure TDP:TOPDEC_PARSING
(**)  structure BA:BASIS
(**)  structure EX:EXECUTION
(**)
(**)  val current_basis : BA.Basis ref

    end

    type topdec

    type SigId
    type StrId
    type LongStrId
    type FunId
 
    type SpecElt
    type Spec
    type Sig
    type CodeElt
    type Code
    type Str
    type StrBody
    type Fun

    type SigEnv
    type StrEnv
    type FunEnv
    type SubSigEnv

    type AllEnv

    type Premise
    type Conclusion
    type Depth
    type Obligation
    type Obligations

    type basis

    exception PVNotInTE
    exception PVNotInSE
    exception PVNotInFE

    val empty_Sp: Spec
    val empty_Co: Code
    val empty_SSE: SubSigEnv
    val empty_T: Sig
    val empty_SB: StrBody
    val empty_S: Str

    val empty_TE: SigEnv
    val empty_SE: StrEnv
    val empty_FE: FunEnv
    val init_A: AllEnv


    val TE_of_A: AllEnv -> SigEnv
    val SB_of_A: AllEnv -> StrBody
    val SE_of_A: AllEnv -> StrEnv
    val FE_of_A: AllEnv -> FunEnv

    val get_Sig: SigId * AllEnv -> Sig
    val get_Str: StrId * AllEnv -> Str
    val get_Fun: FunId * AllEnv -> Fun

    val add_Sig: SigId * Sig * SigEnv -> SigEnv
    val add_Str: StrId * Str * StrEnv -> StrEnv
    val add_Fun: FunId * Fun * FunEnv -> FunEnv

    val make_StrId: string -> StrId
    val make_SigId: string -> SigId
    val make_FunId: string -> FunId

    val T_of_S: Str -> Sig
    val SB_of_S: Str -> StrBody
    val breakup_S: Str -> Sig * StrBody
    val construct_S: Sig * StrBody -> Str

    val Tin_of_F: Fun -> Sig
    val SN_of_F: Fun -> StrId
    val Tout_of_F: Fun -> Sig
    val SB_of_F: Fun -> StrBody
    val breakup_F: Fun -> Sig * StrId * Sig * StrBody
    val construct_F: Sig * StrId * Sig * StrBody -> Fun

    val Co_of_SB: StrBody -> Code
    val SE_of_SB: StrBody -> StrEnv
    val breakup_SB: StrBody -> Code * StrEnv
    val construct_SB: Code * StrEnv -> StrBody

    val Sp_of_T: Sig -> Spec
    val SSE_of_T: Sig -> SubSigEnv
    val breakup_T: Sig -> Spec * SubSigEnv
    val construct_T: Spec * SubSigEnv -> Sig

    val TE_in_A: SigEnv -> AllEnv
    val SE_in_A: StrEnv -> AllEnv
    val SB_in_A: StrBody -> AllEnv
    val FE_in_A: FunEnv -> AllEnv
    val Co_in_SB: Code -> StrBody
    val Sp_in_T: Spec -> Sig
    val SSE_in_T: SubSigEnv -> Sig
    val SE_in_SB: StrEnv -> StrBody
    val CoE_to_Co: CodeElt -> Code
    val SpE_to_Sp: SpecElt -> Spec

    val Sp_plus_Sp: Spec * Spec -> Spec
    val Co_plus_Co: Code * Code -> Code

    val SB_plus_SB: StrBody * StrBody -> StrBody
    val T_plus_T: Sig * Sig -> Sig

    val TE_plus_TE: SigEnv * SigEnv -> SigEnv
    val SE_plus_SE: StrEnv * StrEnv -> StrEnv
    val SSE_plus_SSE: SubSigEnv * SubSigEnv -> SubSigEnv
    val FE_plus_FE: FunEnv * FunEnv -> FunEnv
    val A_plus_A: AllEnv * AllEnv -> AllEnv

    val A_plus_TE: AllEnv * SigEnv -> AllEnv
    val A_plus_SB: AllEnv * StrBody -> AllEnv
    val A_plus_SE: AllEnv * StrEnv -> AllEnv
    val A_plus_FE: AllEnv * FunEnv -> AllEnv
    val S_plus_S: Str * Str -> Str
    
    val single_SSE: StrId * Sig -> SubSigEnv
    val single_SE: StrId * Str -> StrEnv
    val single_TE: SigId * Sig -> SigEnv
    val single_FE: FunId * Fun -> FunEnv

    val in_Dom_SE: StrId * StrEnv -> bool
    val in_Dom_SSE: StrId * SubSigEnv -> bool
    val in_Dom_FE: FunId * FunEnv -> bool
    val is_SB_Undefined: StrBody -> bool


    val remove_Info: 'a OK.TDP.PreElabDecGrammar.WithInfo -> 'a

    val remove_Info_from_LSNlist: (LongStrId OK.TDP.PreElabDecGrammar.WithInfo) list ->
                                  LongStrId list

    val remove_Info_from_TNlist: (SigId OK.TDP.PreElabDecGrammar.WithInfo) list ->
                                 SigId list

    val lookup_TN_in_TE: SigId * SigEnv -> Sig
    val lookup_SN_in_SE: StrId * StrEnv -> Str
    val lookup_FN_in_FE: FunId * FunEnv -> Fun

    val T_from_LSNlist: (LongStrId list) * StrEnv -> Sig

    val T_to_S: Sig -> Str
    val T_to_SB: Sig -> StrBody
    val A_plus_SSE: AllEnv * SubSigEnv -> AllEnv
    val SB_from_LSNlist: (LongStrId list) * StrEnv -> StrBody

    val empty_Os: Obligations
    val Os_plus_O: Obligations * Obligation -> Obligations
    val SB_to_Pr: StrBody -> Premise
    val T_to_Pr: Sig -> Premise
    val SB_to_Cc: StrBody -> Conclusion
    val T_to_Cc: Sig -> Conclusion
    val empty_deep: Depth
    val fun_deep: FunId -> Depth
    val str_deeper: Depth * StrId -> Depth
    val construct_O: Depth * Premise * Conclusion -> Obligation

    val simplify_O: Obligation -> Obligation

    val results: Obligations ref
    val add_result: Obligation -> unit
    val clear_results: unit
    val get_string_results: Obligations -> string
    val show_me_results: Obligations -> unit


    val new_A_from_topdec: topdec * AllEnv -> AllEnv
    val parse_file_to_topdec: string * basis -> topdec
    val deal_with_file: string * AllEnv -> AllEnv
    val cur_base: basis ref 

end

functor PVEnv(K:sig
      val parse: unit -> unit
      val elab: unit -> unit
      val eval: unit -> unit

      val parseFile: string -> unit
      val elabFile: string -> unit
      val evalFile: string -> unit

(* PVADD *)
(**)  structure TDP:TOPDEC_PARSING
(**)  sharing type TDP.Basics.Ident.longid = TDP.PreElabDecGrammar.longid 
(**)  structure BA:BASIS
(**)    sharing type BA.InfixBasis = TDP.Parse.InfixBasis 
(**)  structure EX:EXECUTION
(**)
(**)  val current_basis : BA.Basis ref

    end
):PVENV =
  struct

    structure OK = 
      K
    structure Topdecparse = 
      OK.TDP
    structure Topdecgrammar = 
      Topdecparse.PreElabTopdecGrammar
    structure PPTopdecgrammar = 
      Topdecparse.PPPreElabTopdecGrammar
    structure Decgrammar = 
      Topdecparse.PreElabDecGrammar
    structure PPDecgrammar = 
      Topdecparse.PPPreElabDecGrammar
    structure Basics = 
      Topdecparse.Basics
    structure SigId = 
      Basics.SigId
    structure StrId = 
      Basics.StrId
    structure FunId =
      Basics.FunId
    structure Ident =
      Basics.Ident
    structure Finmap = 
      Basics.Tools.FinMap
    structure PP = 
      Basics.Tools.PrettyPrint
    structure Basis = 
      OK.BA

    type topdec = 
      Topdecgrammar.topdec

    type BaseAx =
      Decgrammar.exp

    type ValBind =
      Decgrammar.valbind

    type DatBind =
      Decgrammar.datbind

    type TypBind =
      Decgrammar.typbind

    type ExBind =
      Decgrammar.exbind

    type ValDesc =
      Topdecgrammar.valdesc

    type DatDesc =
      Topdecgrammar.datdesc

    type TypDesc =
      Topdecgrammar.typdesc

    type ExDesc =
      Topdecgrammar.exdesc

    type SharEq =
      Topdecgrammar.shareq

    type SigId = 
      SigId.sigid

    type StrId = 
      StrId.strid

    type LongStrId =
      StrId.longstrid

    type FunId =
      FunId.funid

    datatype SpecElt =
      LOCALspec of Sig * Sig
    | BASEspec of BaseAx
    | VALspec of ValDesc
    | SPANTYPEspec of DatDesc
    | DATATYPEspec of DatDesc
    | TYPEspec of TypDesc
    | EQTYPEspec of TypDesc
    | EXCEPTIONspec of ExDesc
    | SHARINGspec of SharEq

    and Spec =
      SPEC of SpecElt list

    and Sig =
      SIG of Spec * SubSigEnv

    and SubSigEnv =
      SUBSIGENV of (StrId,Sig) Finmap.map

    datatype SigEnv =
      SIGENV of (SigId,Sig) Finmap.map

    datatype CodeElt =
      LOCALcode of StrBody * StrBody
    | BASEcode of BaseAx
    | VALcode of ValBind
    | SPANTYPEcode of DatBind
    | DATATYPEcode of DatBind
    | TYPEcode of TypBind
    | EQTYPEcode of TypBind
    | EXCEPTIONcode of ExBind

    and Code =
      CODE of CodeElt list

    and Str =
      STR of Sig * StrBody

    and StrBody =
      STRBODY of Code * StrEnv
    | UNDEFINED

    and StrEnv =
      STRENV of (StrId,Str) Finmap.map

    datatype Fun =
      FUN of Sig * StrId * Sig * StrBody

    datatype FunEnv =
      FUNENV of (FunId,Fun) Finmap.map

    datatype AllEnv =
      ALLENV of SigEnv * StrBody * FunEnv

    datatype Premise =
      STRBODYpremise of StrBody
    | SIGpremise of Sig

    datatype Conclusion =
      STRBODYconclusion of StrBody
    | SIGconclusion of Sig

    datatype Depth =
      NODEPTH
    | STRDEPTH of StrId list
    | FUNDEPTH of FunId * (StrId list)

    datatype Obligation =
      OBLIGATION of Depth * Premise * Conclusion

    datatype Obligations =
      OBLIGATIONS of Obligation list

    type string_tree =
      PPTopdecgrammar.StringTree   

    type basis =
      Basis.Basis

    exception PVNotInTE
    exception PVNotInSE
    exception PVNotInFE

    val empty_Sp =
      SPEC []

    val empty_Co =
      CODE []

    val empty_SSE =
      SUBSIGENV Finmap.empty

    val empty_T =
      SIG(empty_Sp,empty_SSE)

    val empty_SE =
      STRENV Finmap.empty

    val empty_SB =
      STRBODY(empty_Co,empty_SE)

    val empty_S =
      STR(empty_T,empty_SB)

    val empty_FE =
      FUNENV Finmap.empty

    val empty_TE =
      SIGENV Finmap.empty

    val init_A =
      ALLENV(empty_TE,empty_SB,empty_FE)

    fun TE_of_A(ALLENV(TE,_,_)) =
      TE

    fun SB_of_A(ALLENV(_,SB,_)) =
      SB

    fun SE_of_A(ALLENV(_,STRBODY(_,SE),_)) =
      SE
    |   SE_of_A(ALLENV(_,UNDEFINED,_)) =
      empty_SE

    fun FE_of_A(ALLENV(_,_,FE)) =
      FE

    fun get_SigMap(SIGENV(TEmap)) =
      TEmap

    fun get_StrMap(STRENV(SEmap)) =
      SEmap

    fun get_FunMap(FUNENV(FEmap)) =
      FEmap

    fun get_SubSigMap(SUBSIGENV(SSEmap)) =
      SSEmap

    fun deal_with_Sig_result (Some T) =
      T
    |   deal_with_Sig_result None =
      raise PVNotInTE

    fun get_Sig(TN,A) =
      deal_with_Sig_result (Finmap.lookup (get_SigMap(TE_of_A A)) TN)

    fun deal_with_Str_result (Some S) =
      S
    |   deal_with_Str_result None =
      raise PVNotInSE

    fun get_Str(SN,A) =
      deal_with_Str_result (Finmap.lookup (get_StrMap(SE_of_A A)) SN)

    fun deal_with_Fun_result (Some F) =
      F
    |   deal_with_Fun_result None =
      raise PVNotInFE

    fun get_Fun(FN,A) =
      deal_with_Fun_result (Finmap.lookup (get_FunMap(FE_of_A A)) FN)


    fun add_Sig(TN,T,TE) =
      SIGENV(Finmap.add(TN,T,(get_SigMap TE)))

    fun add_Str(SN,S,SE) =
      STRENV(Finmap.add(SN,S,(get_StrMap SE)))

    fun add_Fun(FN,F,FE) =
      FUNENV(Finmap.add(FN,F,(get_FunMap FE)))

    val make_StrId =
      StrId.mk_StrId
    
    val make_SigId = 
      SigId.mk_SigId

    val make_FunId =
      FunId.mk_FunId

    fun T_of_S(STR(T,_)) =
      T

    fun SB_of_S(STR(_,SB)) =
      SB

    fun breakup_S(STR(T,SB)) =
      (T,SB)

    fun construct_S(T,SB) =
      STR(T,SB)

    fun Tin_of_F(FUN(T,_,_,_)) =
      T

    fun SN_of_F(FUN(_,SN,_,_)) =
      SN

    fun Tout_of_F(FUN(_,_,T,_)) =
      T

    fun SB_of_F(FUN(_,_,_,SB)) =
      SB

    fun breakup_F(FUN(T,SN,T',SB)) =
      (T,SN,T',SB)

    fun construct_F(T,SN,T',SB) =
      FUN(T,SN,T',SB)

    fun Co_of_SB(STRBODY(Co,_)) =
      Co
    |   Co_of_SB(UNDEFINED) =
      empty_Co

    fun SE_of_SB(STRBODY(_,SE)) =
      SE
    |   SE_of_SB(UNDEFINED) =
      empty_SE

    fun breakup_SB(STRBODY(Co,SE)) =
      (Co,SE)
    |   breakup_SB(UNDEFINED) =
      (empty_Co,empty_SE)

    fun construct_SB(Co,SE) =
      STRBODY(Co,SE)

    fun Sp_of_T(SIG(Sp,_)) =
      Sp

    fun SSE_of_T(SIG(_,SSE)) =
      SSE

    fun breakup_T(SIG(Sp,SSE)) =
      (Sp,SSE)

    fun construct_T(Sp,SSE) =
      SIG(Sp,SSE)

    fun Co_in_SB Co =
      STRBODY(Co,empty_SE)
 
    fun Sp_in_T Sp =
      SIG(Sp,empty_SSE)

    fun SSE_in_T SSE =
      SIG(empty_Sp,SSE)
      
    fun SE_in_SB SE =
      STRBODY(empty_Co,SE)

    fun CoE_to_Co CoE =
      CODE [CoE]

    fun SpE_to_Sp SpE =
      SPEC [SpE]

    fun TE_in_A TE =
      ALLENV(TE,empty_SB,empty_FE)

    fun SB_in_A SB =
      ALLENV(empty_TE,SB,empty_FE)

    fun SE_in_A SE =
      ALLENV(empty_TE,(SE_in_SB SE),empty_FE)

    fun FE_in_A FE =
      ALLENV(empty_TE,empty_SB,FE)

    fun Sp_plus_Sp(SPEC(Sp1),SPEC(Sp2)) =
      SPEC(Sp1@Sp2)

    fun Co_plus_Co(CODE(Co1),CODE(Co2)) =
      CODE(Co1@Co2)

    fun TE_plus_TE(TE1,TE2) =
      SIGENV(Finmap.plus((get_SigMap TE1),(get_SigMap TE2)))

    fun SE_plus_SE(SE1,SE2) =
      STRENV(Finmap.plus((get_StrMap SE1),(get_StrMap SE2)))

    fun SSE_plus_SSE(SSE1,SSE2) =
      SUBSIGENV(Finmap.plus((get_SubSigMap SSE1),(get_SubSigMap SSE2)))

    fun FE_plus_FE(FE1,FE2) =
      FUNENV(Finmap.plus((get_FunMap FE1),(get_FunMap FE2)))

    fun SB_plus_SB(SB1,SB2) =
      let
        val (Co1,SE1) =
          breakup_SB SB1
        val (Co2,SE2) =
          breakup_SB SB2
      in
        STRBODY((Co_plus_Co(Co1,Co2)),(SE_plus_SE(SE1,SE2)))
      end

    fun T_plus_T(T1,T2) =
      let
        val (Sp1,SSE1) =
          breakup_T T1
        val (Sp2,SSE2) =
          breakup_T T2
      in
        SIG((Sp_plus_Sp(Sp1,Sp2)),(SSE_plus_SSE(SSE1,SSE2)))
      end

    fun S_plus_S(S1,S2) =
      let
        val (T1,SB1) =
          breakup_S S1
        val (T2,SB2) =
          breakup_S S2
      in
        STR((T_plus_T(T1,T2)),(SB_plus_SB(SB1,SB2)))
      end

    fun A_plus_A(A1,A2) =
      ALLENV(TE_plus_TE((TE_of_A A1),
                        (TE_of_A A2)),
             SB_plus_SB((SB_of_A A1),
                        (SB_of_A A2)),
             FE_plus_FE((FE_of_A A1),
                        (FE_of_A A2)))

    fun A_plus_TE(A,TE) =
      ALLENV(TE_plus_TE((TE_of_A A),TE),
             (SB_of_A A),
             (FE_of_A A))

    fun A_plus_SB(A,SB) =
      ALLENV((TE_of_A A),
             SB_plus_SB((SB_of_A A),SB),
             (FE_of_A A))

    fun A_plus_SE(A,SE) =
      A_plus_SB(A,(SB_plus_SB((SB_of_A A),(SE_in_SB SE))))

    fun A_plus_FE(A,FE) =
      ALLENV((TE_of_A A),
             (SB_of_A A),
             FE_plus_FE((FE_of_A A),FE))

    fun single_SSE(SN,T) =
      SUBSIGENV(Finmap.singleton(SN,T))

    fun single_SE(SN,S) =
      STRENV(Finmap.singleton(SN,S))

    fun single_TE(SN,T) =
      SIGENV(Finmap.singleton(SN,T))

    fun single_FE(FN,F) =
      FUNENV(Finmap.singleton(FN,F))

    fun Some_or_None(Some(_)) =
      true
    |   Some_or_None(None) =
      false

    fun in_Dom_SE(SN,SE) =
      Some_or_None(Finmap.lookup (get_StrMap SE) SN)

    fun in_Dom_SSE(SN,SSE) =
      Some_or_None(Finmap.lookup (get_SubSigMap SSE) SN)

    fun in_Dom_FE(FN,FE) =
      Some_or_None(Finmap.lookup (get_FunMap FE) FN)

    fun is_SB_Undefined UNDEFINED =
      true
    |   is_SB_Undefined _ =
      false

    fun remove_Info(Decgrammar.WITH_INFO(_,obj)) =
      obj

    fun remove_Info_from_LSNlist [] =
      []
    |   remove_Info_from_LSNlist(LSNi::LSNiS) =
      (remove_Info LSNi)::(remove_Info_from_LSNlist LSNiS)

    fun remove_Info_from_TNlist [] =
      []
    |   remove_Info_from_TNlist(TNi::TNiS) =
      (remove_Info TNi)::(remove_Info_from_TNlist TNiS)

    fun lookup_TN_in_TE(TN,TE) =
      deal_with_Sig_result(Finmap.lookup (get_SigMap TE) TN)

    fun lookup_SN_in_SE(SN,SE) =
      deal_with_Str_result(Finmap.lookup (get_StrMap SE) SN)

    fun lookup_FN_in_FE(FN,FE) =
      deal_with_Fun_result(Finmap.lookup (get_FunMap FE) FN)

    fun lookup_SN_in_SSE(SN,SSE) =
     deal_with_Sig_result(Finmap.lookup (get_SubSigMap SSE) SN)

    fun T_from_TNlist([],TE) =
      empty_T
    |   T_from_TNlist((TN::TNS),TE) =
      T_plus_T((lookup_TN_in_TE(TN,TE)),(T_from_TNlist(TNS,TE)))

    fun SNlist_from_LSN LSN =
      let
        val (strids,strid) =
          StrId.explode_longstrid LSN
      in
        strids@[strid]
      end

    fun T_of_SNlist_in_SSE([SN],SSE) =
      lookup_SN_in_SSE(SN,SSE)
    |   T_of_SNlist_in_SSE((SN::SNs),SSE) =
      T_of_SNlist_in_SSE(SNs,(SSE_of_T(lookup_SN_in_SSE(SN,SSE))))

    fun T_of_SNlist([SN],SE) =
      T_of_S(lookup_SN_in_SE(SN,SE))
    |   T_of_SNlist((SN::SNs),SE) =
      T_of_SNlist_in_SSE(SNs,(SSE_of_T(T_of_S(lookup_SN_in_SE(SN,SE)))))

    fun T_from_LSN(LSN,SE) =
      T_of_SNlist((SNlist_from_LSN LSN),SE)

    fun T_from_LSNlist([],_) =
      empty_T
    |   T_from_LSNlist((LSN::LSNs),SE) =
      T_plus_T((T_from_LSN(LSN,SE)),(T_from_LSNlist(LSNs,SE)))

    fun op_opt obj =
      Decgrammar.OP_OPT(obj,false)

    fun VD_to_VB(Topdecgrammar.VALDESC(i,id,ty,None)) =
      Decgrammar.UNDEFvalbind(i,Decgrammar.TYPEDpat(i,
        Decgrammar.ATPATpat(i,
          Decgrammar.LONGIDatpat(i,(op_opt(Ident.idToLongId id)))),ty),None)
    |   VD_to_VB(Topdecgrammar.VALDESC(i,id,ty,Some(more))) =
      let
        val VB =
          VD_to_VB more
      in
        Decgrammar.UNDEFvalbind(i,Decgrammar.TYPEDpat(i,
          Decgrammar.ATPATpat(i,
            Decgrammar.LONGIDatpat(i,(op_opt(Ident.idToLongId id)))),ty),Some(VB))
      end

    fun condesc_to_conbind(Topdecgrammar.CONDESC(i,con,tyoption,None)) =
      Decgrammar.CONBIND(i,(op_opt con),tyoption,None)
    |   condesc_to_conbind(Topdecgrammar.CONDESC(i,con,tyoption,Some(more))) =
      let
        val conbind =
          condesc_to_conbind more
      in
        Decgrammar.CONBIND(i,(op_opt con),tyoption,Some(conbind))
      end

    fun DD_to_DB(Topdecgrammar.DATDESC(i,tyvarlist,tycon,condesc,None)) =
      Decgrammar.DATBIND(i,tyvarlist,tycon,
            (condesc_to_conbind condesc),None)
    |   DD_to_DB(Topdecgrammar.DATDESC(i,tyvarlist,tycon,condesc,Some(more))) =
      let
        val datbind =
          DD_to_DB more
      in
        Decgrammar.DATBIND(i,tyvarlist,tycon,
            (condesc_to_conbind condesc),Some(datbind))
      end

    fun TD_to_TB(Topdecgrammar.TYPDESC(i,tyvarlist,tycon,None)) =
      Decgrammar.UNDEFtypbind(i,tyvarlist,tycon,None)
    |   TD_to_TB(Topdecgrammar.TYPDESC(i,tyvarlist,tycon,Some(more))) =
      let
        val typbind =
          TD_to_TB more
      in
        Decgrammar.UNDEFtypbind(i,tyvarlist,tycon,Some(typbind))
      end

    fun ED_to_EB(Topdecgrammar.EXDESC(i,excon,tyoption,None)) =
      Decgrammar.EXBIND(i,(op_opt excon),tyoption,None)
    |   ED_to_EB(Topdecgrammar.EXDESC(i,excon,tyoption,Some(more))) =
      let
        val exbind =
          ED_to_EB more
      in
        Decgrammar.EXBIND(i,(op_opt excon),tyoption,Some(exbind))
      end

    fun SpE_to_CoE(BASEspec(BA)) =
      BASEcode BA
    |   SpE_to_CoE(VALspec(VD)) =
      VALcode(VD_to_VB VD) 
    |   SpE_to_CoE(DATATYPEspec(DD)) =
      DATATYPEcode(DD_to_DB DD)
    |   SpE_to_CoE(SPANTYPEspec(DD)) =
      SPANTYPEcode(DD_to_DB DD)
    |   SpE_to_CoE(TYPEspec(TD)) =
      TYPEcode(TD_to_TB TD)
    |   SpE_to_CoE(EQTYPEspec(TD)) =
      EQTYPEcode(TD_to_TB TD)
    |   SpE_to_CoE(EXCEPTIONspec(ED)) =
      EXCEPTIONcode(ED_to_EB ED)
    |   SpE_to_CoE(LOCALspec(T1,T2)) =
      LOCALcode((T_to_SB T1),(T_to_SB T2))

    and Sp_to_Co(SPEC[]) =
      CODE[]
    |   Sp_to_Co(SPEC((SHARINGspec(_))::SpEs)) =
      (Sp_to_Co(SPEC(SpEs)))
    |   Sp_to_Co(SPEC(SpE::SpEs)) =
      Co_plus_Co((CoE_to_Co (SpE_to_CoE SpE)),(Sp_to_Co(SPEC(SpEs))))

    and SSE_to_SE SSE =
      let
        val SSEmap =
          get_SubSigMap SSE
        val SEmap =
          Finmap.composemap (T_to_S) SSEmap
      in
        STRENV SEmap
      end

    and T_to_S T =
      STR(T,UNDEFINED)

    and T_to_SB(SIG(Sp,SSE)) =
      STRBODY((Sp_to_Co Sp),(SSE_to_SE SSE))

    fun T_to_no_SB_S T =
      STR(T,UNDEFINED)

    fun SSE_to_no_SB_SE SSE =
      let
        val SSEmap =
          get_SubSigMap SSE
        val SEmap =
          Finmap.composemap (T_to_no_SB_S) SSEmap
      in
        STRENV SEmap
      end

    fun A_plus_SSE(A,SSE) =
      A_plus_SE(A,(SSE_to_no_SB_SE SSE))

    fun SB_from_LSN(LSN,SE) =
      T_to_SB(T_from_LSN(LSN,SE))

    fun SB_from_LSNlist([],_) =
      empty_SB
    |   SB_from_LSNlist((LSN::LSNs),SE) =
      SB_plus_SB((SB_from_LSN(LSN,SE)),(SB_from_LSNlist(LSNs,SE)))

    fun SB_to_Pr SB =
      STRBODYpremise SB

    fun T_to_Pr T =
      SIGpremise T

    fun SB_to_Cc SB =
      STRBODYconclusion SB

    fun T_to_Cc T =
      SIGconclusion T

    val empty_deep =
      NODEPTH

    fun fun_deep FN =
      FUNDEPTH(FN,[])

    fun str_deeper(NODEPTH,SN) =
      STRDEPTH [SN]
    |   str_deeper(STRDEPTH(SNlist),SN) =
      STRDEPTH(SNlist@[SN])
    |   str_deeper(FUNDEPTH(FN,SNlist),SN) =
      FUNDEPTH(FN,(SNlist@[SN]))

    fun construct_O(D,Pr,Cc) =
      OBLIGATION(D,Pr,Cc)

    val eq_BaseAx  =
      Decgrammar.eq_exp

    val eq_ValBind =
      Decgrammar.eq_valbind

    val eq_DatBind =
      Decgrammar.eq_datbind

    val eq_TypBind =
      Decgrammar.eq_typbind

    val eq_ExBind =
      Decgrammar.eq_exbind

    fun is_SpE_in_SpE(BASEspec(B1),BASEspec(B2)) =
      eq_BaseAx(B1,B2)
    |   is_SpE_in_SpE(SpE,LOCALspec(T1,T2)) =
      is_SpE_in_T(SpE,T2)
    |   is_SpE_in_SpE(_,_) =
      false

    and is_SpE_in_Sp(SpE,(SPEC[])) =
      false
    |   is_SpE_in_Sp(SpE,(SPEC(SpE'::SpEs))) =
      (is_SpE_in_SpE(SpE,SpE')) orelse
      (is_SpE_in_Sp(SpE,(SPEC SpEs)))

    and is_SpE_in_T(SpE,(SIG(Sp,_))) =
      is_SpE_in_Sp(SpE,Sp)

    fun is_SpE_in_CoE(BASEspec(B1),BASEcode(B2)) =
      eq_BaseAx(B1,B2)
    |   is_SpE_in_CoE(SpE,LOCALcode(_,SB2)) =
      is_SpE_in_SB(SpE,SB2)
    |   is_SpE_in_CoE(_,_) =
      false

    and is_SpE_in_Co(SpE,(CODE[])) =
      false
    |   is_SpE_in_Co(SpE,(CODE(CoE::CoEs))) =
      (is_SpE_in_CoE(SpE,CoE)) orelse
      (is_SpE_in_Co(SpE,(CODE CoEs)))

    and is_SpE_in_SB(SpE,(STRBODY(Co,_))) =
      is_SpE_in_Co(SpE,Co)

    fun is_CoE_in_CoE(BASEcode(B1),BASEcode(B2)) =
      eq_BaseAx(B1,B2)
    |   is_CoE_in_CoE(CoE,LOCALcode(SB1,SB2)) =
      is_CoE_in_SB(CoE,SB2)
    |   is_CoE_in_CoE(VALcode(VB1),VALcode(VB2)) =
      eq_ValBind(VB1,VB2)
    |   is_CoE_in_CoE(SPANTYPEcode(DB1),SPANTYPEcode(DB2)) =
      eq_DatBind(DB1,DB2)
    |   is_CoE_in_CoE(DATATYPEcode(DB1),DATATYPEcode(DB2)) =
      eq_DatBind(DB1,DB2)
    |   is_CoE_in_CoE(TYPEcode(TB1),TYPEcode(TB2)) =
      eq_TypBind(TB1,TB2)
    |   is_CoE_in_CoE(EQTYPEcode(TB1),EQTYPEcode(TB2)) =
      eq_TypBind(TB1,TB2)
    |   is_CoE_in_CoE(EXCEPTIONcode(EB1),EXCEPTIONcode(EB2)) =
      eq_ExBind(EB1,EB2)
    |   is_CoE_in_CoE(_,_) =
      false

    and is_CoE_in_Co(CoE,(CODE[])) =
      false
    |   is_CoE_in_Co(CoE,(CODE(CoE'::CoEs))) =
      (is_CoE_in_CoE(CoE,CoE')) orelse
      (is_CoE_in_Co(CoE,(CODE CoEs)))

    and is_CoE_in_SB(CoE,(STRBODY(Co,_))) =
      is_CoE_in_Co(CoE,Co)

    fun is_CoE_in_Pr(CoE,(STRBODYpremise SB)) =
      is_CoE_in_SB(CoE,SB)
    |   is_CoE_in_Pr(CoE,(SIGpremise T)) =
      false

    fun is_SpE_in_Pr(SpE,(STRBODYpremise SB)) =
      is_SpE_in_SB(SpE,SB)
    |   is_SpE_in_Pr(SpE,(SIGpremise T)) =
      is_SpE_in_T(SpE,T)

    fun simplify_SN_by_SSE SSE2 (SN1,T1) =
      if in_Dom_SSE(SN1,SSE2) then
        simplify_T_by_Pr(T1,SIGpremise(lookup_SN_in_SSE(SN1,SSE2)))
      else
        T1

    and simplify_SSE_by_SSE(SSE1,SSE2) =
      Finmap.ComposeMap (simplify_SN_by_SSE SSE2) (get_SubSigMap SSE1)

    and simplify_SN_by_SE SE2 (SN1,T1) =
      if in_Dom_SE(SN1,SE2) then
        simplify_T_by_Pr(T1,SIGpremise((T_of_S(lookup_SN_in_SE(SN1,SE2)))))
      else
        T1

    and simplify_SSE_by_SE(SSE1,SE2) =
      Finmap.ComposeMap (simplify_SN_by_SE SE2) (get_SubSigMap SSE1)
      
    and simplify_SSE_by_Pr(SSE,(STRBODYpremise SB)) =
      SUBSIGENV(simplify_SSE_by_SE(SSE,(SE_of_SB SB)))
    |   simplify_SSE_by_Pr(SSE,(SIGpremise T)) =
      SUBSIGENV(simplify_SSE_by_SSE(SSE,(SSE_of_T T)))

    and simplify_CoE_by_Pr(LOCALcode(SB1,SB2),Pr) =
      (LOCALcode(SB1,(simplify_SB_by_Pr(SB2,Pr))),false)
    |   simplify_CoE_by_Pr(CoE,Pr) =
      (CoE,(is_CoE_in_Pr(CoE,Pr)))
        
    and simplify_Co_by_Pr((CODE []),Pr) =
      CODE []
    |   simplify_Co_by_Pr((CODE(CoE::CoEs)),Pr) =
      let
        val (obj,remove) =
          simplify_CoE_by_Pr(CoE,Pr)
      in
        if remove then
          simplify_Co_by_Pr((CODE CoEs),Pr)
        else
          Co_plus_Co((CODE[obj]),(simplify_Co_by_Pr((CODE CoEs),Pr)))
      end

    and simplify_SB_by_Pr(STRBODY(Co,SE),Pr) =
      let
        val simp_Co =
          simplify_Co_by_Pr(Co,Pr)
      in
        STRBODY(simp_Co,SE)
      end
    |   simplify_SB_by_Pr(UNDEFINED,Pr) =
      UNDEFINED

    and simplify_SpE_by_Pr(LOCALspec(T1,T2),Pr) =
      (LOCALspec(T1,(simplify_T_by_Pr(T2,Pr))),false)
    |   simplify_SpE_by_Pr(SpE,Pr) =
      (SpE,(is_SpE_in_Pr(SpE,Pr)))

    and simplify_Sp_by_Pr((SPEC []),Pr) =
      SPEC []
    |   simplify_Sp_by_Pr((SPEC(SpE::SpEs)),Pr) =
      let
        val (obj,remove) =
          simplify_SpE_by_Pr(SpE,Pr)
      in
        if remove then
          simplify_Sp_by_Pr((SPEC SpEs),Pr)
        else
          Sp_plus_Sp((SPEC[obj]),(simplify_Sp_by_Pr((SPEC SpEs),Pr)))
      end

    and simplify_T_by_Pr(SIG(Sp,SSE),Pr) =
      let
        val simp_Sp =
          simplify_Sp_by_Pr(Sp,Pr)
        val simp_SSE =
          simplify_SSE_by_Pr(SSE,Pr)
      in
        SIG(simp_Sp,simp_SSE)
      end

    fun simplify_Cc_by_Pr(STRBODYconclusion(SB),Pr) =
      STRBODYconclusion(simplify_SB_by_Pr(SB,Pr))
    |   simplify_Cc_by_Pr(SIGconclusion(T),Pr) =
      SIGconclusion(simplify_T_by_Pr(T,Pr))
   
    fun simplify_O(OBLIGATION(D,Pr,Cc)) =
      OBLIGATION(D,Pr,(simplify_Cc_by_Pr(Cc,Pr)))

    val empty_Os =
      OBLIGATIONS []

    fun Os_plus_O(OBLIGATIONS(Os),O) =
      let
        val O' =
          simplify_O O
      in
        OBLIGATIONS(Os@[O'])
      end

    val results =
      ref empty_Os

    fun add_result O =
      results :=
        Os_plus_O((!results),O)

    val clear_results =
      results :=
        empty_Os

    fun Co_from_ax(Decgrammar.AXIOMS(_,exp,None)) =
      CODE [BASEcode exp]
    |   Co_from_ax(Decgrammar.AXIOMS(_,exp,Some(more))) =
      let
        val Co =
          Co_from_ax(more)
      in
        Co_plus_Co((CODE [BASEcode exp]),Co)
      end

    fun Co_from_valdec(VB) =
      CODE [VALcode VB]

    fun Co_from_datatypedec(DB) =
      CODE [DATATYPEcode DB]

    fun Co_from_spantypedec(DB) =
      CODE [SPANTYPEcode DB]

    fun Co_from_typedec(TB) =
      CODE [TYPEcode TB]

    fun Co_from_eqtypedec(TB) =
      CODE [EQTYPEcode TB]

    fun Co_from_exceptiondec(EB) =
      CODE [EXCEPTIONcode EB]

    fun SB_from_dec(Decgrammar.VALdec(_,valbind),SE) =
      Co_in_SB(Co_from_valdec valbind)
    |   SB_from_dec(Decgrammar.SPANTYPEdec(_,datbind),SE) =
      Co_in_SB(Co_from_spantypedec datbind)
    |   SB_from_dec(Decgrammar.DATATYPEdec(_,datbind),SE) =
      Co_in_SB(Co_from_datatypedec datbind)
    |   SB_from_dec(Decgrammar.ABSTYPEdec(_,datbind,dec),SE) =
      SB_from_dec(dec,SE)
    |   SB_from_dec(Decgrammar.TYPEdec(_,typbind),SE) =
      Co_in_SB(Co_from_typedec typbind)
    |   SB_from_dec(Decgrammar.EQTYPEdec(_,typbind),SE) =
      Co_in_SB(Co_from_eqtypedec typbind)
    |   SB_from_dec(Decgrammar.EXCEPTIONdec(_,exbind),SE) =
      Co_in_SB(Co_from_exceptiondec exbind)
    |   SB_from_dec(Decgrammar.AXIOMdec(_,ax),SE) =
      Co_in_SB(Co_from_ax ax)
    |   SB_from_dec(Decgrammar.LOCALdec(_,dec1,dec2),SE) =
      let
        val SB1 =
          SB_from_dec(dec1,SE)
        val SE1 =
          SE_of_SB SB1
        val SB2 =
          SB_from_dec(dec2,(SE_plus_SE(SE,SE1)))
        val SE2 =
          SE_of_SB SB2
      in
        construct_SB((CoE_to_Co (LOCALcode(SB1,SB2))),SE2)
      end
    |   SB_from_dec(Decgrammar.OPENdec(_,LSNilist),SE) =
      SB_from_LSNlist((remove_Info_from_LSNlist LSNilist),SE)
    |   SB_from_dec(Decgrammar.SEQdec(_,dec1,dec2),SE) =
      let
        val SB1 =
          SB_from_dec(dec1,SE)
        val SE1 =
          SE_of_SB SB1
        val SB2 =
          SB_from_dec(dec2,(SE_plus_SE(SE,SE1)))
      in
        SB_plus_SB(SB1,SB2)
      end
    |   SB_from_dec _ =
      empty_SB

    fun Sp_from_ax(Decgrammar.AXIOMS(_,exp,None)) =
      SPEC [BASEspec exp]
    |   Sp_from_ax(Decgrammar.AXIOMS(_,exp,Some(more))) =
      let
        val Sp =
          Sp_from_ax(more)
      in
        Sp_plus_Sp((SpE_to_Sp(BASEspec exp)),Sp)
      end

    fun Sp_from_valdesc(VD) =
      SPEC [VALspec VD]

    fun Sp_from_datatypedesc(DD) =
      SPEC [DATATYPEspec DD]

    fun Sp_from_spantypedesc(DD) =
      SPEC [SPANTYPEspec DD]

    fun Sp_of_Typedesc(TD) =
      SPEC [TYPEspec TD]

    fun Sp_from_eqtypedesc(TD) =
      SPEC [EQTYPEspec TD]

    fun Sp_from_exceptiondesc(ED) =
      SPEC [EXCEPTIONspec ED]

    fun T_of_Sigid(TN,TE) =
      lookup_TN_in_TE(TN,TE)

    fun SSE_from_strdesc(Topdecgrammar.STRDESC(_,strid,sigexp,None),A) =
      let
        val T =
          T_of_Sigexp(sigexp,A)
      in
        single_SSE(strid,T)
      end
    |   SSE_from_strdesc(Topdecgrammar.STRDESC(_,strid,sigexp,Some(more)),A) =
      let
        val SSE =
          SSE_from_strdesc(more,A)
        val T =
          T_of_Sigexp(sigexp,A)
      in
        SSE_plus_SSE((single_SSE(strid,T)),SSE)
      end

    and T_of_Spec(Topdecgrammar.AXIOMspec(_,ax),_) =
      let
        val Sp =
          Sp_from_ax ax
      in
        Sp_in_T Sp
      end
    |   T_of_Spec(Topdecgrammar.VALspec(_,valdesc),_) =
      Sp_in_T(Sp_from_valdesc valdesc)
    |   T_of_Spec(Topdecgrammar.TYPEspec(_,typedesc),_) =
      Sp_in_T(Sp_of_Typedesc typedesc)
    |   T_of_Spec(Topdecgrammar.EQTYPEspec(_,eqtypedesc),_) =
      Sp_in_T(Sp_from_eqtypedesc eqtypedesc)
    |   T_of_Spec(Topdecgrammar.DATATYPEspec(_,datatypedesc),_) =
      Sp_in_T(Sp_from_datatypedesc datatypedesc)
    |   T_of_Spec(Topdecgrammar.SPANTYPEspec(_,spantypedesc),_) =
      Sp_in_T(Sp_from_spantypedesc spantypedesc)
    |   T_of_Spec(Topdecgrammar.EXCEPTIONspec(_,exceptiondesc),_) =
      Sp_in_T(Sp_from_exceptiondesc exceptiondesc)
    |   T_of_Spec(Topdecgrammar.STRUCTUREspec(_,strdesc),A) =
      let
        val SSE =
          SSE_from_strdesc(strdesc,A)
      in
        SSE_in_T SSE
      end
    |   T_of_Spec(Topdecgrammar.OPENspec(_,LSNilist),A) =
      T_from_LSNlist((remove_Info_from_LSNlist LSNilist),(SE_of_A A))
    |   T_of_Spec(Topdecgrammar.INCLUDEspec(_,TNilist),A) =
      T_from_TNlist((remove_Info_from_TNlist TNilist),(TE_of_A A))
    |   T_of_Spec(Topdecgrammar.SEQspec(_,spec1,spec2),A) =
      let
        val T1 =
          T_of_Spec(spec1,A)
        val SSE1 =
          SSE_of_T T1
        val T2 =
          T_of_Spec(spec2,(A_plus_SSE(A,SSE1)))
      in
        T_plus_T(T1,T2)
      end
    |   T_of_Spec(Topdecgrammar.LOCALspec(_,spec1,spec2),A) =
      let
        val T1 =
          T_of_Spec(spec1,A)
        val SSE1 =
          SSE_of_T T1
        val T2 =
          T_of_Spec(spec2,(A_plus_SSE(A,SSE1)))
        val SSE2 =
          SSE_of_T T2
      in
        construct_T((SpE_to_Sp(LOCALspec(T1,T2))),SSE2)
      end
    |   T_of_Spec _ =
      empty_T

    and T_of_Sigexp(Topdecgrammar.SIGsigexp(_,spec),A) =
      T_of_Spec(spec,A)
    |   T_of_Sigexp(Topdecgrammar.SIGIDsigexp(_,sigid),A) =
      T_of_Sigid(sigid,(TE_of_A A))

    fun TE_from_sigbind(Topdecgrammar.SIGBIND(_,sigid,sigexp,None),A) =
      let
        val T =
          T_of_Sigexp(sigexp,A)
      in
        single_TE(sigid,T)
      end
    |   TE_from_sigbind(Topdecgrammar.SIGBIND(_,sigid,sigexp,Some(more)),A) =
      let
        val TE =
          TE_from_sigbind(more,A)
        val T =
          T_of_Sigexp(sigexp,A)
      in
        TE_plus_TE((single_TE(sigid,T)),TE)
      end

    fun TE_from_sigdec(Topdecgrammar.SIGNATUREsigdec(_,sigbind),A) =
      TE_from_sigbind(sigbind,A)
    |   TE_from_sigdec(Topdecgrammar.SEQsigdec(_,sigdec1,sigdec2),A) =
      let
        val TE1 =
          TE_from_sigdec(sigdec1,A)
        val TE2 =
          TE_from_sigdec(sigdec2,A_plus_TE(A,TE1))
      in
        TE_plus_TE(TE1,TE2)
      end
    |   TE_from_sigdec _ =
      empty_TE

    fun SE_from_strbind(Topdecgrammar.STRBIND(_,strid,Some(sigexp),
                        strexp,None),A,SE,D) =
      if (in_Dom_SE(strid,SE)) then
        if (is_SB_Undefined(SB_of_S(lookup_SN_in_SE(strid,SE)))) then
          let
            val So =
              lookup_SN_in_SE(strid,SE)
            val (To,_) =
              breakup_S So
            val T =
              T_of_Sigexp(sigexp,A)
            val SB =
              SB_from_strexp(strexp,A,empty_SE,(str_deeper(D,strid)))
            val O1 =
              construct_O(D,(T_to_Pr T),(T_to_Cc To))
            val O2 =
              construct_O(D,(SB_to_Pr SB),(T_to_Cc T))
          in
            add_result O1;
            add_result O2;
            single_SE(strid,(construct_S(T,SB)))
          end
        else
          let
            val So =
              lookup_SN_in_SE(strid,SE)
            val (To,SBo) =
              breakup_S So
            val T =
              T_of_Sigexp(sigexp,A)
            val SB =
              SB_from_strexp(strexp,A,(SE_of_SB SBo),(str_deeper(D,strid)))
            val O1 =
              construct_O(D,(SB_to_Pr SB),(SB_to_Cc SBo))
            val O2 =
              construct_O(D,(T_to_Pr T),(T_to_Cc To))
            val O3 =
              construct_O(D,(SB_to_Pr SB),(T_to_Cc T))
          in
            add_result O1;
            add_result O2;
            add_result O3;
            single_SE(strid,(construct_S(T,SB)))
          end
      else
        let
          val T =
            T_of_Sigexp(sigexp,A)
          val SB =
            SB_from_strexp(strexp,A,empty_SE,(str_deeper(D,strid)))
          val O1 =
            construct_O(D,(SB_to_Pr SB),(T_to_Cc T))
        in
          add_result O1;
          single_SE(strid,(construct_S(T,SB)))
        end
    |   SE_from_strbind(Topdecgrammar.STRBIND(_,strid,Some(sigexp),
                        strexp,Some(more)),A,SE,D) =
      if (in_Dom_SE(strid,SE)) then
        if (is_SB_Undefined(SB_of_S(lookup_SN_in_SE(strid,SE)))) then
          let
            val SE' =
              SE_from_strbind(more,A,SE,D)
            val So =
              lookup_SN_in_SE(strid,SE)
            val (To,_) =
              breakup_S So
            val T =
              T_of_Sigexp(sigexp,A)
            val SB =
              SB_from_strexp(strexp,A,empty_SE,(str_deeper(D,strid)))
            val O1 =
              construct_O(D,(T_to_Pr T),(T_to_Cc To))
            val O2 =
              construct_O(D,(SB_to_Pr SB),(T_to_Cc T))
          in
            add_result O1;
            add_result O2;
            SE_plus_SE(single_SE(strid,construct_S(T,SB)),SE')
          end
        else
          let
            val SE' =
              SE_from_strbind(more,A,SE,D)
            val So =
              lookup_SN_in_SE(strid,SE)
            val (To,SBo) =
              breakup_S So
            val T =
              T_of_Sigexp(sigexp,A)
            val SB =
              SB_from_strexp(strexp,A,(SE_of_SB SBo),(str_deeper(D,strid)))
            val O1 =
              construct_O(D,(SB_to_Pr SB),(SB_to_Cc SBo))
            val O2 =
              construct_O(D,(T_to_Pr T),(T_to_Cc To))
            val O3 =
              construct_O(D,(SB_to_Pr SB),(T_to_Cc T))
          in
            add_result O1;
            add_result O2;
            add_result O3;
            SE_plus_SE(single_SE(strid,construct_S(T,SB)),SE')
          end
      else
        let
          val SE' =
            SE_from_strbind(more,A,SE,D)
          val T =
            T_of_Sigexp(sigexp,A)
          val SB =
            SB_from_strexp(strexp,A,empty_SE,(str_deeper(D,strid)))
          val O1 =
            construct_O(D,(SB_to_Pr SB),(T_to_Cc T))
        in
          add_result O1;
          SE_plus_SE(single_SE(strid,(construct_S(T,SB))),SE')
        end
    |   SE_from_strbind(Topdecgrammar.UNDEFstrbind(_,strid,Some(sigexp),None),A,SE,D) =
      let
        val T =
          T_of_Sigexp(sigexp,A)
      in
        single_SE(strid,construct_S(T,UNDEFINED))
      end
    |   SE_from_strbind(Topdecgrammar.UNDEFstrbind(_,strid,Some(sigexp),Some(more)),A,SE,D) =
      let
        val SE' =
          SE_from_strbind(more,A,SE,D)
        val T =
          T_of_Sigexp(sigexp,A)
      in
        SE_plus_SE(single_SE(strid,construct_S(T,UNDEFINED)),SE')
      end

    and SB_from_strdec(Topdecgrammar.DECstrdec(_,dec),A,SE,D) =
      SB_from_dec(dec,(SE_of_A A))
    |   SB_from_strdec(Topdecgrammar.STRUCTUREstrdec(_,strbind),A,SE,D) =
      SE_in_SB(SE_from_strbind(strbind,A,SE,D))
    |   SB_from_strdec(Topdecgrammar.LOCALstrdec(_,strdec1,strdec2),A,SE,D) =
      let
        val SB1 =
          SB_from_strdec(strdec1,A,SE,D)
        val SE1 =
          SE_of_SB SB1
        val SB2 =
          SB_from_strdec(strdec2,(A_plus_SE(A,SE1)),SE,D)
        val SE2 =
          SE_of_SB SB2
      in
        construct_SB((CoE_to_Co(LOCALcode(SB1,SB2))),SE2)
      end
    |   SB_from_strdec(Topdecgrammar.SEQstrdec(_,strdec1,strdec2),A,SE,D) =
      let
        val SB1 =
          SB_from_strdec(strdec1,A,SE,D)
        val (Co1,SE1) =
          breakup_SB SB1
        val SB2 =
          SB_from_strdec(strdec2,(A_plus_SE(A,SE1)),SE,D)
      in
        SB_plus_SB(SB1,SB2)
      end
    |   SB_from_strdec _ =
      empty_SB

    and SB_from_strexp(Topdecgrammar.STRUCTstrexp(_,strdec),A,SE,D) =
      SB_from_strdec(strdec,A,SE,D)
    |   SB_from_strexp(Topdecgrammar.LONGSTRIDstrexp(_,longstrid),A,SE,D) =
      SB_from_LSN(longstrid,(SE_of_A A))
    |   SB_from_strexp(Topdecgrammar.APPstrexp(_,funid,strexp),A,SE,D) =
      let
        val F =
          lookup_FN_in_FE(funid,(FE_of_A A))
        val (T,_,T',_) =
          breakup_F F
        val SB =
          SB_from_strexp(strexp,A,SE,D)
        val O1 =
          construct_O(D,(SB_to_Pr SB),(T_to_Cc T))
      in
        add_result O1;
	T_to_SB T'
      end
    |   SB_from_strexp(Topdecgrammar.LETstrexp(_,strdec,strexp),A,SE,D) =
      let
        val SB =
          SB_from_strdec(strdec,A,SE,D)
        val SE' =
          SE_of_SB SB
      in
        SB_from_strexp(strexp,(A_plus_SE(A,SE')),SE,D)
      end

    fun FE_from_funbind(Topdecgrammar.FUNBIND(_,funid,strid,sigexp,
                        Some(sigexp'),strexp,None),A,FE,D) =
      if (in_Dom_FE(funid,FE)) then
        if (is_SB_Undefined(SB_of_F(lookup_FN_in_FE(funid,FE)))) then
          let
            val Fo =
              lookup_FN_in_FE(funid,FE)
            val (To,strido,T'o,_) =
              breakup_F Fo
            val T =
              T_of_Sigexp(sigexp,A)
            val Atemp =
              A_plus_SE(A,(single_SE(strid,construct_S(T,UNDEFINED))))
            val T' =
              T_of_Sigexp(sigexp',Atemp)
            val SB =
              SB_from_strexp(strexp,Atemp,empty_SE,(fun_deep funid))
            val O1 =
              construct_O(D,(SB_to_Pr SB),(T_to_Cc T'))
            val O2 =
              construct_O(D,(T_to_Pr T'),(T_to_Cc T'o))
            val O3 =
              construct_O(D,(T_to_Pr To),(T_to_Cc T))
          in
            add_result O1;
            add_result O2;
            add_result O3;
            single_FE(funid,construct_F(T,strid,T',SB))
          end
        else
          let
            val Fo =
              lookup_FN_in_FE(funid,FE)
            val (To,strido,T'o,SBo) =
              breakup_F Fo
            val T =
              T_of_Sigexp(sigexp,A)
            val Atemp =
              A_plus_SE(A,(single_SE(strid,construct_S(T,UNDEFINED))))
            val T' =
              T_of_Sigexp(sigexp',Atemp)
            val SB =
              SB_from_strexp(strexp,Atemp,(SE_of_SB SBo),(fun_deep funid))
            val O1 =
              construct_O(D,(SB_to_Pr SB),(T_to_Cc T'))
            val O2 =
              construct_O(D,(T_to_Pr T'),(T_to_Cc T'o))
            val O3 =
              construct_O(D,(T_to_Pr To),(T_to_Cc T))
            val O4 =
              construct_O(D,(SB_to_Pr SB),(SB_to_Cc SBo))
          in
            add_result O1;
            add_result O2;
            add_result O3;
            add_result O4;
            single_FE(funid,construct_F(T,strid,T',SB))
          end
      else
        let
          val T =
            T_of_Sigexp(sigexp,A)
          val Atemp =
            A_plus_SE(A,(single_SE(strid,construct_S(T,UNDEFINED))))
          val T' =
            T_of_Sigexp(sigexp',Atemp)
          val SB =
            SB_from_strexp(strexp,Atemp,empty_SE,(fun_deep funid))
          val O1 =
            construct_O(D,(SB_to_Pr SB),(T_to_Cc T'))
        in
          add_result O1;
          single_FE(funid,construct_F(T,strid,T',SB))
        end
    |   FE_from_funbind(Topdecgrammar.FUNBIND(_,funid,strid,sigexp,
                        Some(sigexp'),strexp,Some(more)),A,FE,D) =
      if (in_Dom_FE(funid,FE)) then
        if (is_SB_Undefined(SB_of_F(lookup_FN_in_FE(funid,FE)))) then
          let
            val FE' =
              FE_from_funbind(more,A,FE,D)
            val Fo =
              lookup_FN_in_FE(funid,FE)
            val (To,strido,T'o,_) =
              breakup_F Fo
            val T =
              T_of_Sigexp(sigexp,A)
            val Atemp =
              A_plus_SE(A,(single_SE(strid,construct_S(T,UNDEFINED))))
            val T' =
              T_of_Sigexp(sigexp',Atemp)
            val SB =
              SB_from_strexp(strexp,Atemp,empty_SE,(fun_deep funid))
            val O1 =
              construct_O(D,(SB_to_Pr SB),(T_to_Cc T'))
            val O2 =
              construct_O(D,(T_to_Pr T'),(T_to_Cc T'o))
            val O3 =
              construct_O(D,(T_to_Pr T),(T_to_Cc To))
          in
            add_result O1;
            add_result O2;
            add_result O3;
            FE_plus_FE(single_FE(funid,construct_F(T,strid,T',SB)),FE')
          end
        else
          let
            val FE' =
              FE_from_funbind(more,A,FE,D)
            val Fo =
              lookup_FN_in_FE(funid,FE)
            val (To,strido,T'o,SBo) =
              breakup_F Fo
            val T =
              T_of_Sigexp(sigexp,A)
            val Atemp =
              A_plus_SE(A,(single_SE(strid,construct_S(T,UNDEFINED))))
            val T' =
              T_of_Sigexp(sigexp',Atemp)
            val SB =
              SB_from_strexp(strexp,Atemp,(SE_of_SB SBo),(fun_deep funid))
            val O1 =
              construct_O(D,(SB_to_Pr SB),(T_to_Cc T'))
            val O2 =
              construct_O(D,(T_to_Pr T'),(T_to_Cc T'o))
            val O3 =
              construct_O(D,(T_to_Pr T),(T_to_Cc To))
            val O4 =
              construct_O(D,(SB_to_Pr SB),(SB_to_Cc SBo))
          in
            add_result O1;
            add_result O2;
            add_result O3;
            add_result O4;
            FE_plus_FE(single_FE(funid,construct_F(T,strid,T',SB)),FE')
          end
      else
        let
          val FE' =
            FE_from_funbind(more,A,FE,D)
          val T =
            T_of_Sigexp(sigexp,A)
          val Atemp =
            A_plus_SE(A,(single_SE(strid,construct_S(T,UNDEFINED))))
          val T' =
            T_of_Sigexp(sigexp',Atemp)
          val SB =
            SB_from_strexp(strexp,Atemp,empty_SE,(fun_deep funid))
          val O1 =
            construct_O(D,(SB_to_Pr SB),(T_to_Cc T'))
        in
          add_result O1;
          FE_plus_FE(single_FE(funid,construct_F(T,strid,T',SB)),FE')
        end
    |   FE_from_funbind(Topdecgrammar.UNDEFfunbind(_,funid,strid,sigexp,
                        Some(sigexp'),None),A,FE,D) =
      let
        val T =
          T_of_Sigexp(sigexp,A)
        val Atemp =
          A_plus_SE(A,single_SE(strid,(construct_S(T,UNDEFINED))))
        val T' =
          T_of_Sigexp(sigexp',Atemp)
      in
        single_FE(funid,construct_F(T,strid,T',UNDEFINED))
      end				    
    |   FE_from_funbind(Topdecgrammar.UNDEFfunbind(_,funid,strid,sigexp,
                        Some(sigexp'),Some(more)),A,FE,D) =
      let
        val FE' =
          FE_from_funbind(more,A,FE,D)
        val T =
          T_of_Sigexp(sigexp,A)
        val Atemp =
          A_plus_SE(A,single_SE(strid,(construct_S(T,UNDEFINED))))
        val T' =
          T_of_Sigexp(sigexp',Atemp)
      in
        FE_plus_FE(single_FE(funid,construct_F(T,strid,T',UNDEFINED)),FE')
      end				    

    and FE_from_fundec(Topdecgrammar.FUNCTORfundec(_,funbind),A,FE,D) =
      FE_from_funbind(funbind,A,FE,D)
    |   FE_from_fundec(Topdecgrammar.SEQfundec(_,fundec1,fundec2),A,FE,D) =
      let
        val FE1 =
          FE_from_fundec(fundec1,A,FE,D)
        val FE2 =
          FE_from_fundec(fundec2,(A_plus_FE(A,FE1)),FE,D)
      in
        FE_plus_FE(FE1,FE2)
      end
    |   FE_from_fundec _ =
      empty_FE

    fun A_from_topdec(Topdecgrammar.SIGtopdec(_,sigdec),A) =
      TE_in_A(TE_from_sigdec(sigdec,A))
    |   A_from_topdec(Topdecgrammar.STRtopdec(_,strdec),A) =
      SB_in_A(SB_from_strdec(strdec,A,(SE_of_A A),empty_deep))
    |   A_from_topdec(Topdecgrammar.FUNtopdec(_,fundec),A) =
      FE_in_A(FE_from_fundec(fundec,A,(FE_of_A A),empty_deep))  

    fun new_A_from_topdec(topdec,A) =
      A_plus_A(A,(A_from_topdec(topdec,A)))




















    datatype Show =
      STRING of string
    | ONE of int * Show
    | TWO of int * Show * Show
    | THREE of int * Show * Show * Show
    | FOUR of int * Show * Show * Show * Show
    | FIVE of int * Show * Show * Show * Show * Show
    | SEQ of Show * Show
    | LIST of (Show list)

    fun spaces 0 =
      ""
    |   spaces n =
      " "^(spaces(n-1))

    fun indent(n,str) =
      (spaces n)^str

    fun string_show(STRING(str),d) =
      indent(d,str)
    |   string_show(ONE(i,s1),d) =
      string_show(s1,(i+d))^"\n"
    |   string_show(TWO(i,s1,s2),d) =
      (string_show(s1,d))^"\n"^
      (string_show(s2,(i+d)))
    |   string_show(THREE(i,s1,s2,s3),d) = 
      (string_show(s1,d))^"\n"^
      (string_show(s2,(i+d)))^"\n"^
      (string_show(s3,d))
    |   string_show(FOUR(i,s1,s2,s3,s4),d) = 
      (string_show(s1,d))^"\n"^
      (string_show(s2,(i+d)))^"\n"^
      (string_show(s3,d))^"\n"^
      (string_show(s4,(i+d)))
    |   string_show(FIVE(i,s1,s2,s3,s4,s5),d) = 
      (string_show(s1,d))^"\n"^
      (string_show(s2,(i+d)))^"\n"^
      (string_show(s3,d))^"\n"^
      (string_show(s4,(i+d)))^"\n"^
      (string_show(s5,d))
    |   string_show(SEQ(s1,s2),d) =
      (string_show(s1,d))^"\n"^
      (string_show(s2,d))
    |   string_show((LIST []),d) =
      ""
    |   string_show((LIST [s]),d) =
      string_show(s,d)
    |   string_show((LIST(s1::ss)),d) =
      (string_show(s1,d))^"\n"^
      (string_show((LIST ss),d))

    fun show_show Sh =
      print (string_show(Sh,0))

    fun Show_B exp =
      STRING "axiom"

    fun Show_VALcode vb =
      STRING "val"

    fun Show_SPANTYPEcode db =
      STRING "spantype"

    fun Show_DATATYPEcode db =
      STRING "datatype"

    fun Show_EQTYPEcode tb =
      STRING "eqtype"

    fun Show_TYPEcode tb =
      STRING "type"

    fun Show_EXCEPTIONcode eb =
      STRING "exception"

    fun Show_VALspec vd =
      STRING "val"

    fun Show_SPANTYPEspec dd =
      STRING "spantype"

    fun Show_DATATYPEspec dd =
      STRING "datatype"

    fun Show_EQTYPEspec td =
      STRING "eqtype"

    fun Show_TYPEspec td =
      STRING "type"

    fun Show_EXCEPTIONspec ed =
      STRING "exception"

    fun string_SN SN =
      StrId.pr_StrId SN

    fun Show_SN SN =
      STRING (string_SN SN)

    fun string_SNlist [] =
      ""
    |   string_SNlist (SN::SNs) =
      (string_SN SN)^"."^(string_SNlist SNs)

    fun Show_LSN LSN =
      STRING (string_SNlist (SNlist_from_LSN LSN))

    fun string_FN FN =
      FunId.pr_FunId FN

    fun Show_FN FN =
      STRING (string_FN FN)

    fun string_TN TN =
      SigId.pr_SigId TN

    fun Show_TN TN =
      STRING (string_TN TN)

   fun get_rangeelt(Some(rangeelt)) =
     rangeelt

   fun Show_map(showdomelt,showrangeelt,themap) =
      let
        val dom =
          Finmap.domSORTED themap
        fun showelt domelt =
          TWO(2,(showdomelt domelt),(showrangeelt(get_rangeelt(Finmap.lookup themap domelt))))
      in
        LIST(map showelt dom)
      end

    fun Show_SpE(LOCALspec(T1,T2)) =
      FIVE(2,(STRING "local"),(Show_T T1),(STRING "in"),(Show_T T2),(STRING "end"))
    |   Show_SpE(BASEspec B) =
      Show_B B
    |   Show_SpE(VALspec vd) =
      Show_VALspec vd
    |   Show_SpE(SPANTYPEspec dd) =
      Show_SPANTYPEspec dd
    |   Show_SpE(DATATYPEspec dd) =
      Show_DATATYPEspec dd
    |   Show_SpE(EQTYPEspec td) =
      Show_EQTYPEspec td
    |   Show_SpE(TYPEspec td) =
      Show_TYPEspec td
    |   Show_SpE(EXCEPTIONspec ed) =
      Show_EXCEPTIONspec ed

    and Show_Sp(SPEC SpEs) =
      LIST(map Show_SpE SpEs)

    and Show_T(SIG(Sp,SSE)) =
      THREE(2,(STRING "sig"),(SEQ((Show_Sp Sp),(Show_SSE SSE))),(STRING "end"))

    and Show_SSE(SUBSIGENV SSEmap) =
(*      Show_map(Show_SN,Show_T,SSEmap) *)
      STRING "SubSigEnv"

    fun Show_CoE(LOCALcode(SB1,SB2)) =
      FIVE(2,(STRING "local"),(Show_SB SB1),(STRING "in"),(Show_SB SB2),(STRING "end"))
    |   Show_CoE(BASEcode B) =
      Show_B B
    |   Show_CoE(VALcode vd) =
      Show_VALcode vd
    |   Show_CoE(SPANTYPEcode dd) =
      Show_SPANTYPEcode dd
    |   Show_CoE(DATATYPEcode dd) =
      Show_DATATYPEcode dd
    |   Show_CoE(EQTYPEcode td) =
      Show_EQTYPEcode td
    |   Show_CoE(TYPEcode td) =
      Show_TYPEcode td
    |   Show_CoE(EXCEPTIONcode ed) =
      Show_EXCEPTIONcode ed

    and Show_Co(CODE CoEs) =
      LIST(map Show_CoE CoEs)

    and Show_SB(STRBODY(Co,SE)) =
      THREE(2,(STRING "struct"),(SEQ((Show_Co Co),(Show_SE SE))),(STRING "end"))

    and Show_SE(STRENV SEmap) =
(*        Show_map(Show_SN,Show_S,SEmap)  *)
      STRING "StrEnv"

    and Show_S(STR(T,SB)) =
      SEQ(THREE(2,(STRING "sig"),(Show_T T),(STRING "end =")),
          THREE(2,(STRING "struct"),(Show_SB SB),(STRING "end")))

    fun Show_Cc(STRBODYconclusion SB) =
      TWO(2,(STRING "Conclusion"),Show_SB SB)
    |   Show_Cc(SIGconclusion T) =
      TWO(2,(STRING "Conclusion"),Show_T T)

    fun Show_Pr(STRBODYpremise SB) =
      TWO(2,(STRING "Premise"),Show_SB SB)
    |   Show_Pr(SIGpremise T) =
      TWO(2,(STRING "Premise"),Show_T T)

    fun string_D NODEPTH =
      ""
    |   string_D(STRDEPTH SNlist) =
      string_SNlist SNlist
    |   string_D(FUNDEPTH(FN,SNlist)) =
      (string_FN FN)^(string_SNlist SNlist)

    fun Show_D D =
      STRING ("From depth : "^(string_D D))

    fun Show_O(OBLIGATION(D,Pr,Cc)) =
      SEQ((TWO(2,(Show_D D),
              (SEQ((Show_Pr Pr),
                   (Show_Cc Cc))))),
          (STRING "\n"))

    fun Show_Os(OBLIGATIONS Os) =
      LIST(map Show_O Os)









    fun isolate_topdec(Topdecparse.Parse.SUCCESS(_,td,_)) =
      td

    fun parse_file_to_topdec(fname,Basis) =
      isolate_topdec(Topdecparse.Parse.parse((Basis.Inf_of_B Basis),
                     (Topdecparse.Parse.begin(Topdecparse.Parse.sourceFromFile fname))))

    fun get_string_results Os =
      string_show((Show_Os Os),0)

    fun show_me_results Os =
      show_show(Show_Os Os)

    val cur_base = OK.current_basis

    fun deal_with_file (file_name,AllEnv) =
      new_A_from_topdec(parse_file_to_topdec(file_name,(!cur_base)),AllEnv)
  end
  









